#include <qtextstream.h>
#include <qfile.h>
#include <iostream>
#include <vector>

#include <assert.h>
#include <Scene.h>

#include "CoTeXIO.h"

#include "../Textures/SimpleColorTexture.h"
#include "../Textures/PerlinNoiseColorTexture.h"

#include "../SurfelGenerator/SurfelGenerator.h"



using namespace std;

static const unsigned int SAVE_BASE_NUM = 1; 
static const size_t NODE_VECTOR_DEFAULT_SIZE = 1024; // with this, there's no need to enlarge the vector for most of the ctx files

void CoTeXIO::save (TextureInterface::TextureMap * texMap,
		    CoTeXTreeInterface * tree, 
		    const QString & filename) {
  assert (tree != NULL 
	  && texMap != NULL);

  QFile file (filename);
  if (!file.open (IO_WriteOnly))
    throw CoTeXIOException ("Error opening file");

  QTextStream stream (&file);
  
  stream << "TEXTURES" << endl;
  
  TextureInterface::TextureMap::Iterator itTex;
  for (itTex = texMap->begin (); itTex != texMap->end (); itTex++) {
    stream << "1 ";
    (*itTex)->save (stream);
    stream << endl;
  }
  stream << "0 " << endl; // between textures and the Constructive Texturing Tree (CTT)
  
  stream << "CTT" << endl;
  
  if (!tree->isEmpty ()) {
    int base = SAVE_BASE_NUM;
    saveNode (tree->getRoot (), base, stream);
  }
  stream << "0" << endl; // at the end of the file 
  file.close ();
}

void CoTeXIO::load (CoTeXTreeInterface * tree, 
		    Object * activeObject,
		    CoTeXConfigurationWidget * confWidget,
		    PrimitiveManagerInterface::PrimitiveObjectMap * primitiveObjectMap,
		    PrimitiveManagerInterface::PrimitiveManagerMap * primitiveManagerMap,
		    OperatorManagerInterface::OperatorManagerMap * operatorManagerMap,
		    const QString & filename) {
  assert (tree != NULL 
	  && confWidget != NULL 
	  && primitiveObjectMap != NULL
	  && primitiveManagerMap != NULL
	  && operatorManagerMap != NULL);
  
  QFile file (filename);
  if (!file.open (IO_ReadOnly))
    throw CoTeXIOException ("Error opening file");
  
  QTextStream stream (&file);
  QString tmp;
    
  while (tmp != "TEXTURES")
    stream >> tmp ; // reads comments
  
  loadTextures (confWidget, stream);
  
  while (tmp != "CTT")
    stream >> tmp; // reads comments
  
  loadTree (tree, activeObject, confWidget, primitiveObjectMap, primitiveManagerMap,
	    operatorManagerMap, stream);

  file.close();
}

void CoTeXIO::saveNode (CoTeXNode * node, int & num, QTextStream &stream) {
  int myNum, lChildNum = -1, rChildNum = -1;
  
  myNum = num++;
  
  if (node->getLeftChild () != NULL) {
    lChildNum = num;
    saveNode (node->getLeftChild (), num, stream);
  }
  if (node->getRightChild () != NULL) {
    rChildNum = num;
    saveNode (node->getRightChild (), num, stream);
  }
  
  stream << myNum << " " << lChildNum << " " << rChildNum << " ";
  node->save (stream);
  stream << endl; 
}

void CoTeXIO::loadTextures (CoTeXConfigurationWidget * confWidget,
			    QTextStream & stream) {
  QString type;
  int num = 0;
  
  stream >> num ;
  while (num != 0){ 
    stream >> type ;
    
    if (type == "SimpleColorTexture"){
      cout << "simple" << endl ;
      SimpleColorTexture * tex = new SimpleColorTexture ();
      tex->load (stream);
      confWidget->insertTexture (tex);
    }
    else if (type == "PerlinNoiseColorTexture"){
      cout << "perlin" << endl ;
      PerlinNoiseColorTexture * tex = new PerlinNoiseColorTexture ();
      tex->load (stream);
      confWidget->insertTexture (tex);
    } 
    else
      throw CoTeXIOException ("Error not recognize texture type"); 
    
    stream >> num ;
  }
}
  

void CoTeXIO::loadTree (CoTeXTreeInterface * tree,
			Object * activeObject,
			CoTeXConfigurationWidget * confWidget,
			PrimitiveManagerInterface::PrimitiveObjectMap * primitiveObjectMap,
			PrimitiveManagerInterface::PrimitiveManagerMap * primitiveManagerMap,
			OperatorManagerInterface::OperatorManagerMap * operatorManagerMap,
			QTextStream &stream) {
  
  int num;
  QString type, subType, tmp;
  vector<CoTeXNode*> nodeVector (NODE_VECTOR_DEFAULT_SIZE);
  CoTeXNode * node ;
  
  stream >> num ;

  while (num != 0) { // 0  the end of the file
    int lChild, rChild;
    
    stream >> lChild >> rChild >> type >> subType;
    if (type == "operator") {
      if (!operatorManagerMap->contains (subType))
	throw CoTeXIOException ("Bad operator type in file");
      node = (*operatorManagerMap)[subType]->createOperator ();
    }
    else if (type == "primitive") {
      if (!primitiveManagerMap->contains (subType))
	throw CoTeXIOException ("Bad primitive type in file");
      node = (*primitiveManagerMap)[subType]->createPrimitive ();
    }
    else
      throw CoTeXIOException ("Bad type in file");
    
    nodeVector.reserve (num + 1);
    if (num >= (int)nodeVector.size ())
      nodeVector.resize (num + 1, NULL);
    
    nodeVector[num] = node;

    if (lChild != -1) {
      node->setLeftChild (nodeVector[lChild]);
      nodeVector[lChild]->setFather (node);
    }
    else
      node->setLeftChild (NULL);
    if (rChild != -1) {
      node->setRightChild (nodeVector[rChild]);
      nodeVector[rChild]->setFather (node);
    }
    else
      node->setRightChild (NULL);
      
    node->load (stream);
    
    if (type == "primitive") {
      QString texName;
      stream >> texName ;
      PrimitiveInterface * primitive = (PrimitiveInterface*)node;
      primitive->setTexture (confWidget->getTexture (texName));
      Scene * scene = Scene::getInstance ();
      primitiveObjectMap->insert (primitive, generateObject (primitive, activeObject));
      scene->addObject ((*primitiveObjectMap)[primitive], false, false);
    }
    stream >> num ;
  }
  tree->setRoot (node);
}
