#ifndef PRIMITIVEINTERFACE_H
#define PRIMITIVEINTERFACE_H

#include "HyperFunable.h"
#include "CoTeXNode.h"
#include "TextureInterface.h"
#include "../BBoxMinMax/BBoxMinMax.h"


/// Interface for CoTeX Primitives
/**
 * Primitives are CoTeX space partitons : they must give the attribute evaluated at 
 * a specified 3D point.
 */ 
class PrimitiveInterface : public CoTeXNode{
public:
  
  /// The base constructor
  /**
   * All primtives must have a name and a non null texture
   * @param name the name, which can unique with the StringTool 
   *             (but it's manager 's work).
   * @param t the texture, which mustn't be NULL (assertion).
   */
  PrimitiveInterface (QString name, TextureInterface * t) :  CoTeXNode (name), texture (t) {}
  
  
  /// The destructor
  /**
   * There's only one member, the texture, which can be used by several primitives,
   * so it mustn't be destroyed by the primitive itself!
   */ 
  virtual ~PrimitiveInterface () {}

  
  /// True if the specified point is inside the primitive.
  /**
   * @param position the specified position to test.
   */
  inline bool contains (const Vector3D& position) { return (eval (position) >= 0);}

  
  // *********
  // Accessors
  // *********
  inline TextureInterface * getTexture () const { return texture;}
  
  
  /// Set the primitive's texture
  /**
   * @param t a non NULL texture
   */
  void setTexture (TextureInterface * t);

  /// Give a bounding box for the primitive.
  /**
   * It's useful for some algorithms which need to work in a resteind space
   * (such as for polygonization or surfelization).
   */
  virtual BBoxMinMax getBBoxMinMax () = 0;
  
  // *******************
  // Interface CoTeXNode 
  // *******************
  
  
  /// From CoTeXNode : here, primitives are leaves.
  bool check () { return (getLeftChild () == NULL && getRightChild () == NULL);}
  
  
  /// The same for all primitives, so defined here.
  /**
   * All primitives have the same method for evaluation : taking a specified position,
   * {x,y,z} (<code>Vector3D</code>), it returns the value of the carried texture
   * at {x,y,z} if it contains the position, NULL if not.
   */
  virtual const AttributeInterface * getAttribute (const Vector3D& position);

 
private:

  TextureInterface * texture; /// the texture carried by the primitive
};

#endif //PRIMITIVEINTERFACE_H
