#ifndef BLOCKMANAGER_H
#define BLOCKMANAGER_H

#include <qpainter.h>

#include <Object.h>
#include <Vector2D.h>

#include "../Attributes/ColorAttribute.h"
#include "../Interfaces/PrimitiveManagerInterface.h"

#include "Block.h"


/// Events manager for blocks.
/**
 * The BlockManager is able to create blocks, based on the UI events and the 
 * current Scene. A segment, respresenting the size of the block 
 * is update during the creation for having a feedback without 
 * re-rendering (drawn in 2D). The version implemented here
 * produce blocsk with same value for dx, dy and dz (cubes).
 */
class BlockManager : public PrimitiveManagerInterface {
public :

  static BlockManager * getInstance ();
  static void destroyInstance ();

  PrimitiveInterface * getCurrentPrimitive (); 
  Object * getCurrentObject ();
  PrimitiveInterface * createPrimitive (); 

  /**
   * Gets the <code>SurfelInterface</code> corresponding to the clicked position 
   * of the mouse from <code>FrameBufferInterface</code>, and creates an <code>Block</code>
   * with this <code>SurfelInterface</code> as the center of the sphere and dx, dy, dz equal to 0.
   *
   * @param event the UI event gived by CoTeX.
   *
   * @param confWidget the CoTeXConfigurationWidget, 
   *                   for quering options of creation (current texture, etc ...).
   *
   * @see #SurfelInterface
   * @see #FrameBufferInterface
   */
  void handleMousePressEvent (QMouseEvent *event, 
			      CoTeXConfigurationWidget * confWidget);

  /**
   * Handles the event of moving the mouse : update the size of the block
   * @param event the UI event gived by CoTeX.
   */
  void handleMouseMoveEvent (QMouseEvent *event);
  
  /**
   * Define the final size (distance beetween original and current position).
   * The current <code>Sphere</code> is ready to be gived by <code>getInstance ()</code>.
   * An Object representing the block is aviable (<code>getObject ()</code>).
   *
   * @see #SurfelInterface
   * @see #FrameBufferInterface
   */
  void handleMouseReleaseEvent (QMouseEvent *);
  
  /// Not useful for sphere creation
  void handleKeyPressEvent (QKeyEvent *ke);

  /// Not useful for sphere creation
  void handleKeyReleaseEvent (QKeyEvent *ke);
  
protected :

  /// Protected, because of the design pattern <code>Singleton</code>
  BlockManager ();  
  /// Protected, because of the design pattern <code>Singleton</code> 
  virtual ~BlockManager ();

private :
  
  bool mousePressed;
  QPoint originalMousePosition; ///< the screen original center when drawing
  QPoint lastMousePosition; ///< the screen last position when moving mouse
  bool blockVisible;
  float screenSize;
  float distance;
  
  Block * currentPrimitive; ///< the block currently constructed
  Object * currentPrimitiveObject; ///< the PS object representing the current instance
  int instanceCount; ///< for generating unique names.
  
  QPainter painter; ///< for painting visual artefacts.
  Vector3D p3D[8]; ///< height points of the cube in 3D
  Vector2D p2D[8]; ///< height points of the cube in 2D
  
  Vector3D center;
  CoTeXConfigurationWidget * configurationWidget;
  

  /// Destroy current non null resources. 
  void clean ();

  /// Draw a segment, with the same position and size than the current Block.
  void drawBlock ();
};

#endif // BLOCKMANAGER_H
