#ifndef CONVPOINT_H
#define CONVPOINT_H

#include <vector>
#include <Vector3D.h>

#include "../Interfaces/PrimitiveInterface.h"
#include "../Interfaces/TextureInterface.h"
#include "../Textures/SimpleColorTexture.h"

static const double DEFAULT_CONV_THRESHOLD = 0.00001; ///< enough small for good results

/// A ConvPoint, set of connected skeleton line segments.
/**
 * Here, the convolution is described by a skelton, and a common kernel value, 
 * called S. A threshold is present for evaluation.
 */
class ConvPoint : public PrimitiveInterface{
public :
  
  /// Set of points defining the skeleton of a convolution.
  /**
   * This vector store the skeleton 's points, and the order 
   * of these points (the first is at position 0, the second at position 1, etc ...)
   */
  typedef std::vector<Vector3D> ConvSkeleton;

  /// General constructor for ConvPoint with points
  /**
   * @param n
   *        name of the convolution
   * @param commonS
   *        in this implementation, all kernels have the same value for S.
   * @param ske
   *        set of points, defining the skeleton
   * @param t
   *        the texture carried by this primitive.
   * @param thres 
   *        threshold, must be positive.
   */
  ConvPoint (QString n, double commonS, ConvSkeleton * ske , 
	     TextureInterface * tex = DEFAULT_SIMPLECOLOR,
	     double thres = DEFAULT_CONV_THRESHOLD);
  
  ConvPoint (ConvPoint & conv);
  
  /// Base destructor.
  /**
   * The whole skeleton is destroyed.
   */ 
  virtual ~ConvPoint ();
  
  // *********
  // Accessors
  // *********

  /// Accessor to the kernel value.
  inline float getS () const { return S;}
  
  inline double getThreshold () const { return threshold;}
  
  /// Accessor to the skeleton.
  /**
   * @return ConvSkeleton* the skeleton, a set of <code>Vector3D</code>
   * @see #skeleton
   */
  inline ConvSkeleton * getSkeleton () const { return skeleton;}
  
  /// Sets the value of the kernel.
  void setS (float newS); 
  
  /// Sets the value of the threshold.
  /**
   * @param newThreshold
   *	an double which represents the new value of the threshold of the 
   *              <code>ConvPoint</code>, must be positive, because
   *              convolutions have positive values everywhere, and we use 
   *              threshold for having some null values for the surface.
   * 
   * @see #Eval
   */
  void setThreshold (double newThreshold);
  
  /// Sets the skeleton  with new set of points
  void setSkeleton (ConvSkeleton * newSkeleton);
  
  /// Redefinition of the operator = for the class <code>ConvPoint</code>.
  /**
   * Threshold, radius and skeleton are copied.
   * @param ConvPoint
   *         an <code>ConvPoint</code>
   * @return ConvPoint
   */
  ConvPoint& operator=(ConvPoint& conv);
  
  /// Give a bounding box for the convolution.
  BBoxMinMax getBBoxMinMax ();

  // *******************
  // CoTeXNode Interface
  // *******************
  
  /// F-Rep evaluaton of a ConvPoint with curve
  /**
   * Evaluate the value of the convolution function, decreasing exponentially from the 
   * center but everywhere positive. For having the surface (null values), 
   * we substract the threshold of all evaluations. As the function decrease
   * exponentially, we need a very small threshold for having good results.
   * 
   * @return double.
   */
  double eval (const Vector3D& position);
  
  // **********************
  // HyperFunable Interface
  // **********************
  
  void writeHFString (QTextStream & outStream);
  void writeDefArray (QTextStream & outStream);
  void writeDefBody (QTextStream & outStream);

  // *****************
  // Savable Interface
  // *****************
  
  void save (QTextStream & stream);
  void load (QTextStream & stream);
  
private:

  double S; ///< here we choose to give the same kernel value for the whole skeleton.
  double squareS; ///< for optimisation
  double threshold;
  ConvSkeleton * skeleton;
 
  
  // **************************************************************
  // Actually, all the heavy work of eval is done by this fonction.
  // **************************************************************
  
  /// Compute the convolution as a convolution with a set of skeleton points
  /**
   * One of the more simple convolution, but easy to compute. 
   */
  double computePointValue (const Vector3D& position);


  /// Compute the convolution as a convolution with a curve.
  /**
   * A more complex convolution, based on curves.
   */
  double computeCurveValue (const Vector3D& position);
  
};
  

#endif // CONVPOINT_H
