#include "SphereManager.h"

#include <assert.h>
#include <math.h>

#include <RendererManager.h>
#include <Transformer.h>
#include <Scene.h>
#include <StringTools.h>

#include "../SurfelGenerator/SurfelGenerator.h"

static const unsigned int SPHERE_SEQ_LEN = 4; ///< number of digit for sphere number

static SphereManager * instance = NULL;

SphereManager * SphereManager::getInstance () {
  if (instance == NULL)
    instance = new SphereManager;
  return instance;
}
  
void SphereManager::destroyInstance () {
  if (instance != NULL) {
    delete instance;
    instance = NULL;
  }
}    

SphereManager::SphereManager () : mousePressed (false), circleVisible (false), 
				  currentPrimitive (NULL), currentPrimitiveObject (NULL),
				  instanceCount (0) {}

SphereManager::~SphereManager () { 
  clean ();
}

PrimitiveInterface* SphereManager::getCurrentPrimitive () {
  PrimitiveInterface * res = currentPrimitive;
 
  currentPrimitive = NULL;
  return res;
} 

Object * SphereManager::getCurrentObject () {
  Object * res = currentPrimitiveObject;
  
  currentPrimitiveObject = NULL;
  return res;
}

PrimitiveInterface * SphereManager::createPrimitive () {
  QString newName = StringTools::appendSequenceNumber ("Sphere", // easy way to ensure unique names
						       instanceCount++, 
						       SPHERE_SEQ_LEN);
  return new Sphere (newName);  
}


void SphereManager::clean () {
  if (currentPrimitive != NULL)
    delete currentPrimitive;
  if (currentPrimitiveObject != NULL)
    delete currentPrimitiveObject;
}


void SphereManager::handleMousePressEvent (QMouseEvent *event, 
					   CoTeXConfigurationWidget * confWidget) {
  assert (confWidget != NULL);
  
  if (!mousePressed 
      && event->button () == Qt::LeftButton 
      && event->state () == Qt::NoButton) {
    Scene * scene = Scene::getInstance ();
    RendererInterface * renderer = RendererManager::getInstance()->getMainRenderer();
    FrameBufferInterface * frameBuffer;
    Object * activeObject = scene->getActiveObject ();
    SurfelInterface * surfel = NULL;
    Vector3D center;
    
    clean ();
    mousePressed = true;
    renderer->renderFrame (FrameBufferInterface::PERPIXEL_SURFEL 
			   | FrameBufferInterface::PERPIXEL_C_Z_N_W);
    frameBuffer = renderer->getFrameBuffer();
    if (frameBuffer->getObjectAtPixel (event->x(), event->y()) != activeObject) {
      mousePressed = false;
      return; // only surfels from active object are interesting
    }
    surfel = frameBuffer->getSurfel (event->x (), event->y ());
    if (surfel != NULL) {
      // surfels are expressed in the local mark of the object, 
      // and primitives too for application to other object.
      currentCenter = surfel->getPosition ();
      configurationWidget = confWidget;
      originalMousePosition = event->pos ();
      screenRadius = 0.0;
    }
    else
      mousePressed = false;
  }
}


void SphereManager::handleMouseMoveEvent (QMouseEvent * event) {
  if (mousePressed
      && event->button() == Qt::NoButton) {
    if (circleVisible)
      drawCircle ();
    lastMousePosition = event->pos ();
    QPoint diff = lastMousePosition - originalMousePosition;
    screenRadius = sqrt (diff.x () * diff.x () + diff.y () * diff.y ());
    drawCircle ();
    circleVisible = true;
  }
}

void SphereManager::handleMouseReleaseEvent (QMouseEvent *) {
  if (mousePressed) {
    mousePressed = false;
    if (screenRadius == 0.0) {
      currentPrimitive = NULL;
      throw PrimitiveManagerException ("Too Small Radius");
    }
    else {
      Scene * scene = Scene::getInstance ();
      Transformer * transformer = Transformer::getSceneTransformer();  // Returns a global instance of Transformer, which is updated by the RendererManager
      RendererInterface * renderer = RendererManager::getInstance ()->getMainRenderer ();
      FrameBufferInterface * frameBuffer= renderer->getFrameBuffer ();
      Object * activeObject = scene->getActiveObject ();
      Vector3D point1, point2;
      float trueRadius, depth;
      
      // ****************************************
      // Extraction of the true radius (in the 3D 
      // world) from the screen radius.
      // ****************************************
      
      frameBuffer = renderer->getFrameBuffer ();
      depth = frameBuffer->getDepth (originalMousePosition.x (), 
				     originalMousePosition.y ());
      point1 = transformer->viewPortToObject (Vector2D (originalMousePosition.x (),
							originalMousePosition.y ()),
					      depth);
      point2 = transformer->viewPortToObject (Vector2D (lastMousePosition.x (),  
							lastMousePosition.y ()),
					      depth);
      trueRadius = Vector3D::distance (point1, point2);
      
      // unique name  : "Sphere0001", "Sphere0002", ...
      QString newName = StringTools::appendSequenceNumber ("Sphere", 
							   instanceCount++, 
							   SPHERE_SEQ_LEN);
      currentPrimitive = new Sphere (newName, currentCenter, trueRadius, 
				     configurationWidget->getCurrentTexture ());
      
      // ***********************************
      // Creation of the 3D PointShop Object 
      // representing the primitive.
      // ***********************************

      currentPrimitiveObject = new Object (scene->getUniqueName ("CoTeXSphere"));
      
      // PrimitivesObjects, contrary to primitives, must be expressed in the 
      // global mark.
      currentPrimitiveObject->setPosition (currentPrimitive->getCenter () 
					   + activeObject->getPosition ());
      
      scene->addObject (currentPrimitiveObject, false, false);

      ColorAttribute primitiveObjectColor = 
	createColorForViz (currentPrimitive->getTexture (),
			   currentPrimitive->getCenter ()) ;
      generateSphereSurfels (currentPrimitiveObject, trueRadius, primitiveObjectColor);   
      circleVisible = false; // scene re-rendered
    }
  }
}


void SphereManager::handleKeyPressEvent (QKeyEvent *ke) {
  ke->ignore(); // by now, we'll ignore the key event
}


void SphereManager::handleKeyReleaseEvent (QKeyEvent *ke) {
  ke->ignore(); // by now, we'll ignore the key event
}



// ***************************************
// VISUALISATION OF THE SPHERE BY A CIRCLE
// ***************************************

void SphereManager::drawCircle () {
  
  // Draw a circle for simple and low cost vizualisation. 
  painter.begin (RendererManager::getInstance()->getMainRenderer()->getRendererWidget());
  
  // initialisation of the pen for having a good visualization in all cases 
  // (inverted color).
  painter.setPen (QPen (Qt::white, 1, Qt::SolidLine));
  painter.setRasterOp (Qt::XorROP);

  // this ellipse is in fact a circle with the sphere radius, 
  // centered on the sphere center.
  painter.drawEllipse ((int)(originalMousePosition.x () - screenRadius),
		       (int)(originalMousePosition.y () - screenRadius), 
		       (int)(2 * screenRadius), (int)(2 * screenRadius));
  painter.flush ();
  painter.end ();
}
